#!/usr/bin/env python
import argparse
import yaml
from autorelease import ReleaseNoteWriter, AutoreleaseParsingHelper
from autorelease.gh_api4.notes4 import NotesWriter, prs_since_latest_release

def make_parser():
    parser = argparse.ArgumentParser()
    auto_parser = AutoreleaseParsingHelper(parser)
    auto_parser.add_github_parsing()
    auto_parser.add_project_parsing()
    auto_parser.parser.add_argument("--conf", type=str)
    # auto_parser.parser.add_argument("-o", "--output", type=str)
    # auto_parser.parser.add_argument("--since-release", type=str)
    return auto_parser

def main():
    auto_parser = make_parser()
    opts = auto_parser.parse_args()
    with open(opts.conf, mode='r') as f:
        config = yaml.load(f.read(), Loader=yaml.SafeLoader)


    token = None
    if auto_parser.github_user is not None:
        token = auto_parser.github_user.token
    elif tok :

    notes_conf = config['notes']
    category_labels = {
        lab['label']: lab['heading'] for lab in notes_conf['labels']
    }
    topics = {}
    for label in notes_conf['labels']:
        if tops := label.get('topics'):
            topic_dict = {top['label']: top['name'] for top in tops}
            topics[label['label']] = topic_dict

    writer = NotesWriter(
        category_labels=category_labels,
        topics=topics,
        standard_contributors=notes_conf['standard_contributors']
    )
    import pdb; pdb.set_trace()

    new_prs = prs_since_latest_release(
        owner=config['project']['repo_owner'],
        repo=config['project']['repo_name'],
        auth=opts.github_user
    )
    print(writer_write(new_prs))

    writer = ReleaseNoteWriter(config=opts.conf,
                               since_release=opts.since_release,
                               project=opts.project,
                               github_user=opts.github_user)
    writer.write_release_notes(outfile=opts.output)

if __name__ == "__main__":
    main()



import os
import click
import yaml

from autorelease.scripts.vendor import vendor_actions
from autorelease.scripts.check import run_checks
# from autorelease import ReleaseNoteWriter
from autorelease.gh_api4.notes4 import NotesWriter, prs_since_latest_release

import logging


def _find_first_file(pathlist):
    for p in pathlist:
        if os.path.exists(p):
            return p
    return None

def load_yaml(user_input, default_pathlist):
    if user_input is not None:
        dct = yaml.load(user_input, Loader=yaml.FullLoader)
    else:
        filename = _find_first_file(default_pathlist)
        if filename is None:
            raise RuntimeError("No config file could be found")
        with open(filename, mode='r') as f:
            dct = yaml.load(f, Loader=yaml.FullLoader)

    return dct

def load_config(user_input):
    paths = [
        ".autorelease.yml",
        "autorelease.yml",
        os.path.join(".autorelease", "autorelease.yml"),
        os.path.join(".autorelease", "conf.yml"),
    ]
    return load_yaml(user_input, paths)

def load_auth(user_input):
    paths = [
        os.path.join(os.path.expanduser("~"), ".autorelease-auth.yml"),
        os.path.join(".autorelease", "auth.yml"),
        os.path.join(".autorelease", "autorelease-auth.yml"),
        ".autorelease-auth.yml",
        "autorelease-auth.yml",
    ]
    return load_yaml(user_input, paths)


@click.group()
@click.option("--loglevel", type=str, default="WARNING")
def cli(loglevel):
    logging.basicConfig(level=getattr(logging, loglevel))
    logger = logging.getLogger("autorelease")
    logger.setLevel(getattr(logging, loglevel))

@cli.command()
@click.option('--conf', type=click.File('r'))
@click.option('--branch', default=None)
@click.option('--event', default=None)
def check(conf, branch, event):
    dct = load_config(conf)
    run_checks(dct, branch, event)

@cli.command()
@click.option('--conf', type=click.File('r'))
def config(conf):
    from pprint import pprint
    config = load_config(conf)
    pprint(config)

@cli.command()
@click.option('--auth', type=click.File('r'))
def auth(auth):
    from pprint import pprint
    auth = load_auth(auth)
    pprint(auth)


@cli.command()
@click.option('--conf', type=click.File('r'))
@click.option('--auth', type=click.File('r'))
def notes(conf, auth):
    config = load_config(conf)
    github_user = load_auth(auth)
    target_branch = config['repo'].get('dev-branch', 'main')
    notes_conf = config['notes']
    notes_conf.update(github_user)
    notes_conf['project'] = config['project']
    category_labels = {
        lab['label']: lab['heading'] for lab in notes_conf['labels']
    }
    topics = {}
    for label in notes_conf['labels']:
        if tops := label.get('topics'):
            topic_dict = {top['label']: top['name'] for top in tops}
            topics[label['label']] = topic_dict

    writer = NotesWriter(
        category_labels=category_labels,
        topics=topics,
        standard_contributors=notes_conf['standard_contributors']
    )
    new_prs = prs_since_latest_release(
        owner=config['project']['repo_owner'],
        repo=config['project']['repo_name'],
        auth=(None, github_user['github_user']['token']),
        target_branch=target_branch,
    )
    print(writer.write(new_prs))

@click.group()
def vendor():
    pass

@vendor.command()
def actions():
    print("vendoring actions")
    vendor_actions(base_path='.')

cli.add_command(vendor)

if __name__ == "__main__":
    cli()



from typing import NamedTuple
from .query_runner import QueryRunner
from .utils import string_to_datetime
import datetime

class Release(NamedTuple):
    name: str
    tag: str
    draft: bool
    prerelease: bool
    latest: bool
    date: datetime.datetime

    @classmethod
    def from_api(cls, api_release):
        return cls(
            name=api_release['name'],
            tag=api_release["tagName"],
            draft=api_release["isDraft"],
            prerelease=api_release["isPrerelease"],
            latest=api_release["isLatest"],
            date=string_to_datetime(api_release["publishedAt"]),
        )

RELEASES_QUERY = """
{
  repository(name: "$repo_name", owner: "$repo_owner") {
    releases(orderBy: {field: CREATED_AT, direction: DESC}, first: 100) {
      nodes {
        publishedAt
        isLatest
        isPrerelease
        isDraft
        name
        tagName
      }
    }
  }
}
"""

def latest_release(owner, repo, auth):
    # TODO: support paginated releases
    runner = QueryRunner(RELEASES_QUERY, auth)
    result = runner(repo_name=repo, repo_owner=owner)
    api_release_info = result['data']['repository']['releases']['nodes']
    releases = [Release.from_api(rel) for rel in api_release_info]
    claim_latest = [rel for rel in releases if rel.latest]
    assert len(claim_latest) == 1
    return claim_latest[0]



import requests
import string

GITHUB_API_ENDPOINT = "https://api.github.com/graphql"

class QueryRunner:
    def __init__(self, query_template, auth,
                 api_endpoint=GITHUB_API_ENDPOINT):
        self.query_template = string.Template(query_template)
        self.auth = auth
        self.api_endpoint = api_endpoint

    def __call__(self, **kwargs):
        query = self.query_template.substitute(**kwargs)
        return requests.post(self.api_endpoint, json={'query': query},
                             auth=self.auth).json()



import string
try:
    import pathlib
except ImportError:  # py2
    import pathlib2
import pkg_resources
from packaging.version import Version
import autorelease

import git
import re

def extract_github_owner_and_repo(url):
    github_url_re = ".*github.com:(?P<owner>.*)/(?P<repo>.*)"
    m = re.match(github_url_re, url)
    owner = m.group('owner')
    repo = m.group('repo')
    if repo.endswith('.git'):
        repo = repo[:-4]
    return owner, repo


def guess_parent_repository(repo_name='.'):
    """
    Guess the repository to consider as the "main" repo, based on remotes.

    This returns GitHub repo owner and repo name for the 'upstream' remote
    if it exists, otherwise 'origin' if it exists. If neither exists,
    returns None, None.

    Returns
    -------
    owner : str
        owner of the GitHub repo
    repo : str
        name of the GitHub repo
    """
    repo = git.Repo(repo_name)
    parent = None
    try:
        parent = repo.remotes.upstream
    except AttributeError:
        try:
            parent = repo.remotes.origin
        except AttributeError:
            pass

    if parent is None:
        return None, None
    else:
        return extract_github_owner_and_repo(parent.url)


def _get_github_repo(config):
    # TODO: this guessing should be a backup plan; get info from config if
    # available
    owner, repo = guess_parent_repository('.')
    if owner is None or repo is None:
        raise RuntimeError("Unable to determine repository")
    return owner + "/" + repo


def get_substitution_mapping(config=None):
    version = Version(autorelease.version.version).base_version
    return {
        'VERSION': version,
        'GITHUB_REPO': _get_github_repo(config)
    }


def vendor(resources, base_path, relative_target_dir, substitutions):
    for resource in resources:
        orig_loc = pkg_resources.resource_filename('autorelease', resource)
        name = pathlib.Path(orig_loc).name
        target_dir = base_path / relative_target_dir
        target_dir.mkdir(parents=True, exist_ok=True)
        target_loc = base_path / relative_target_dir / name
        # print(f"cp {orig_loc} {target_loc}")
        with open(orig_loc, mode='r') as rfile:
            template = string.Template(rfile.read())

        with open(target_loc, mode='w') as wfile:
            wfile.write(template.substitute(**substitutions))

def vendor_actions(base_path):
    resources = ['autorelease-default-env.sh', 'autorelease-prep.yml',
                 'autorelease-gh-rel.yml', 'autorelease-deploy.yml']
    resources = ['gh_actions_stages/' + res for res in resources]
    target_dir = pathlib.Path('.github/workflows')
    substitutions = get_substitution_mapping()
    vendor(resources, base_path, target_dir, substitutions)




from .query_runner import QueryRunner

import typing
import enum
import datetime

from .utils import string_to_datetime



class PRStatus(enum.Enum):
    OPEN = "open"
    CLOSED = "closed"
    MERGED = "merged"


class PR(typing.NamedTuple):
    number: int
    target: str
    title: str
    status: PRStatus
    author: str
    labels: typing.List[str]
    url: str
    merge_time: typing.Optional[datetime.datetime]

    @classmethod
    def from_api_response(cls, api_pr):
        return cls(
            number=int(api_pr["number"]),
            target=api_pr["baseRefName"],
            title=api_pr["title"],
            status=getattr(PRStatus, api_pr["state"]),
            author=api_pr["author"]["login"],
            labels=[node["name"] for node in api_pr["labels"]["nodes"]],
            url=api_pr["url"],
            merge_time=string_to_datetime(api_pr["mergedAt"]),
        )

PR_QUERY = """
{
  repository(name: "$repo_name", owner: "$repo_owner") {
    pullRequests(
      orderBy: {field: UPDATED_AT, direction: DESC}
      first: 100
      $after
      states: MERGED
      baseRefName: "$target_branch"
    ) {
      nodes {
        author {
          login
        }
        merged
        mergedAt
        number
        title
        headRefName
        closed
        baseRefName
        state
        url
        labels(first: 100) {
          nodes {
            name
          }
          pageInfo {
            endCursor
            hasNextPage
            startCursor
          }
        }
      }
      pageInfo {
        startCursor
        endCursor
        hasNextPage
        hasPreviousPage
      }
    }
  }
}
"""

def graphql_get_all_prs(owner, repo, auth, target_branch):
    # TODO: query needs to take repo and owner
    def extractor(result):
        return result["data"]["repository"]["pullRequests"]["nodes"]

    def next_page_cursor(result):
        info = result["data"]["repository"]["pullRequests"]["pageInfo"]
        next_cursor = info["endCursor"] if info["hasNextPage"] else None
        return next_cursor


    query_runner = QueryRunner(PR_QUERY, auth=auth,
                               api_endpoint="https://api.github.com/graphql")
    extracted_results = []

    # TODO: how to manage nested inner loops?
    # actually.. better choice is to post-process to remove inner
    # paginations: get additional labels for anything with more than 100
    # labels
    default_kwargs = {
        'repo_owner': owner,
        'repo_name': repo,
        'target_branch': target_branch,
    }
    result = query_runner(after="", **default_kwargs)
    extracted_results.extend(extractor(result))
    while cursor := next_page_cursor(result):
        result = query_runner(after=f'after: "{cursor}"', **default_kwargs)
        extracted_results.extend(extractor(result))

    return extracted_results





"""
Writing release notes based on the GitHub v4 (GraphQL) API
"""

import typing
import enum
import datetime

from .pull_requests import PRStatus, PR, graphql_get_all_prs
from .releases import latest_release

import logging
_logger = logging.getLogger(__name__)

def filter_release_prs(all_prs, prev_release_date, target_branch="main"):
    def is_release_pr(pr):
        return (
            pr.status == PRStatus.MERGED
            and pr.merge_time > prev_release_date
            and pr.target == target_branch
        )

    for pr in all_prs:
        _logger.info(f"{pr}")
        if is_release_pr(pr):
            _logger.info(f"Including {pr}")
            yield pr
        else:
            _logger.info("Skipping")


def prs_since_latest_release(owner, repo, auth, target_branch="main"):
    latest = latest_release(owner, repo, auth)
    _logger.info(f"Latest release: {latest}")

    release_date = latest.date
    all_prs = [PR.from_api_response(pr)
               for pr in graphql_get_all_prs(owner, repo, auth,
                                             target_branch)]

    _logger.info(f"Loaded {len(all_prs)} PRs")
    new_prs = list(filter_release_prs(
        all_prs=all_prs,
        prev_release_date=latest.date,
        target_branch=target_branch
    ))
    _logger.info(f"After filtering, found {len(new_prs)} new PRs")
    return new_prs


class PRCategory:
    def __init__(self, label, heading, topics):
        self.label = label
        self.heading = heading
        self.topics = topics
        self.prs = []
        self.topic_prs = {l: [] for l in topics}

    def append(self, pr):
        if topics := set(pr.labels) & set(self.topics):
            for topic in topics:
                self.topic_prs[topic].append(pr)
        else:
            self.prs.append(pr)


class NotesWriter:
    def __init__(self, category_labels, topics, standard_contributors):
        self.category_labels = category_labels
        self.topics = topics
        self.standard_contributors = set(standard_contributors)

    @staticmethod
    def assign_prs_to_categories(prs, categories):
        category_labels = set(categories)
        for pr in prs:
            selected = [categories[label]
                        for label in set(pr.labels) & category_labels]

            if not selected:
                selected = [categories[None]]

            for category in selected:
                category.append(pr)

    def _write_pr_details(self, pr, category_label, topic_label):
        out = f"[#{pr.number}]({pr.url})"
        if pr.author not in self.standard_contributors:
            out += f" @{pr.author}"

        out_labels = [label for label in pr.labels
                      if label not in {category_label, topic_label}]
        if out_labels:
            out += " "
        out += " ".join(f"#{label}" for label in out_labels)
        return out

    def write_single_pr(self, pr, category_label):
        details = self._write_pr_details(pr, category_label, "")
        out = f"* {pr.title} ({details})\n"
        return out

    def write_topic(self, category, topic):
        out = ""
        topic_prs = category.topic_prs[topic]
        topic_text = category.topics[topic]
        if len(topic_prs):
            out += f"* {topic_text} ("
            out += ", ".join(
                self._write_pr_details(pr, category.label, topic)
                for pr in topic_prs
            )
            out += ")\n"

        return out

    def write_category(self, category):
        out = f"## {category.heading}\n\n"
        for pr in category.prs:
            out += self.write_single_pr(pr, category.label)

        for topic in category.topics:
            out += self.write_topic(category, topic)

        return out

    def write(self, prs):
        categories = {
            label: PRCategory(label, heading, self.topics.get(label, {}))
            for label, heading in self.category_labels.items()
        }
        categories[None] = PRCategory(None, "Unlabeled PRs", {})
        self.assign_prs_to_categories(prs, categories)

        out = "\n".join(self.write_category(category)
                        for category in categories.values())
        return out
